/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.editor.name;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.commands.operations.IUndoableOperation;
import org.eclipse.jface.action.Action;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.views.properties.IPropertySource;

import eu.etaxonomy.cdm.model.name.BotanicalName;
import eu.etaxonomy.cdm.model.name.NameRelationship;
import eu.etaxonomy.cdm.model.name.NonViralName;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.name.ZoologicalName;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.TaxonBase;
import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.TaxEditorPlugin;
import eu.etaxonomy.taxeditor.controller.EditorController;
import eu.etaxonomy.taxeditor.controller.GlobalController;
import eu.etaxonomy.taxeditor.editor.GroupedComposite;
import eu.etaxonomy.taxeditor.model.CdmUtil;
import eu.etaxonomy.taxeditor.operations.name.CreateSynonymInNewGroupOperation;
import eu.etaxonomy.taxeditor.propertysheet.name.BotanicalNamePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.name.NonViralNamePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.name.ZoologicalNamePropertySource;

/**
 * Formats an <code>GroupedComposite</code> to display <code>TaxonNameBase</code> elements
 * in a <code>NameViewer</code>.
 * 
 * @author p.ciardelli
 * @created 02.06.2008
 * @version 1.0
 */
public abstract class NameComposite extends GroupedComposite {
	private static final Logger logger = Logger.getLogger(NameComposite.class);

	/**
	 * ************ COMPOSITE TYPES ************
	 */
	public String compositeType;
	public static final String ACCEPTED_TAXON = "accepted_name_composite";
	public static final String HOMOTYPIC_SYNONYM = "homotypic_name_composite";
	public static final String HETEROTYPIC_SYNONYM = "heterotypic_name_composite";
	public static final String MISAPPLIED_NAME = "misappliedname_name_composite";
	public static final String CONCEPTRELATION = "concept_name_comcposite";

	/**
	 * ************ INDENTATIONS ************
	 */
	public static final int ACCEPTED_INDENT = 0;
	public static final int SYNONYM_INDENT = 15;
	public static final int MISAPPLIEDNAME_INDENT = 15;
	public static final int CONCEPT_INDENT = 15;

	/**
	 * ************ FONTS ************
	 */
	public static final Font ACCEPTED_FONT = TaxEditorPlugin.getDefault()
			.getFont(ITaxEditorConstants.ACCEPTED_TAXON_FONT);
	public static final Font SYNONYM_FONT = TaxEditorPlugin.getDefault()
			.getFont(ITaxEditorConstants.SYNONYM_FONT);
	public static final Font MISAPPLIEDNAME_FONT = TaxEditorPlugin.getDefault()
			.getFont(ITaxEditorConstants.MISAPPLIEDNAME_FONT);
	public static final Font CONCEPT_FONT = TaxEditorPlugin.getDefault()
			.getFont(ITaxEditorConstants.CONCEPT_FONT);

	/**
	 * ************ ICONS ************
	 */
	public static final Image ACCEPTED_ICON = TaxEditorPlugin.getDefault()
			.getImage(ITaxEditorConstants.BLACK_SQUARE_ICON);
	public static final Image HOMOTYPIC_SYNONYM_ICON = TaxEditorPlugin
			.getDefault().getImage(ITaxEditorConstants.HOMOTYPIC_SYN_ICON);
	public static final Image HOMOTYPIC_SYNONYM_ORIGINAL_COMBINATION_ICON = TaxEditorPlugin
			.getDefault().getImage(
					ITaxEditorConstants.HOMOTYPIC_SYN_ORIGINAL_ICON);
	public static final Image HETEROTYPIC_SYNONYM_ICON = TaxEditorPlugin
			.getDefault().getImage(ITaxEditorConstants.HETEROTYPIC_SYN_ICON);
	public static final Image HETEROTYPIC_SYNONYM_ORIGINAL_COMBINATION_ICON = TaxEditorPlugin
			.getDefault().getImage(
					ITaxEditorConstants.HETEROTYPIC_SYN_ORIGINAL_ICON);
	public static final Image MISAPPLIEDNAME_ICON = TaxEditorPlugin
			.getDefault().getImage(ITaxEditorConstants.MISAPPLIED_NAME_ICON);
	public static final Image CONCEPT_ICON = TaxEditorPlugin
			.getDefault().getImage(ITaxEditorConstants.CONCEPT_ICON);
	public static final Image AUTONYM_ICON = TaxEditorPlugin.getDefault()
			.getImage(ITaxEditorConstants.AUTONYM_ICON);
	public static final Image BASIONYM_ICON = TaxEditorPlugin.getDefault()
			.getImage(ITaxEditorConstants.BASIONYM_ICON);
	public static final Image MOVE = TaxEditorPlugin.getDefault().getImage(
			ITaxEditorConstants.MOVE_ICON);

	/**
	 * ************ TRANSFORMATIONS ************
	 */
	public static final String ADD_GROUP_BASIONYM = "add_group_basionym";
	public static final String REMOVE_GROUP_BASIONYM = "remove_group_basionym";

	/**
	 * ************ MENU ACTIONS ************
	 */
	public Action CHANGE_TAXON_TO_SYNONYM_ACTION;

	private static final String EMPTY_NAME_PROMPT = "Click to add name";

	/**
	 * Used to turn parser on and off.
	 * 
	 * @see activateParser
	 * @see deactivateParser
	 */
	private boolean isUseParser = false;

	protected boolean isParsing;

	/**
	 * The constructor for a DescriptionElementComposite. Takes a parent Composite on which to
	 * create itself, and an IManagedForm for Composite life cycle methods, i.e.
	 * drawing borders, creating other Composites, creating line wrap support,
	 * etc.
	 * 
	 * @param parent
	 * @param managedForm
	 */
	public NameComposite(Composite parent, IManagedForm managedForm, 
			TaxonBase taxonBase) {
		super(parent, managedForm);

		createTextViewer();
		createBorderSupport();
		createLineWrapSupport();
		createLineBreakListener();
		
		setDraggableControl(textViewer.getRulerControl());
		
		createParser();
		createEmptyViewerPrompt(EMPTY_NAME_PROMPT);
		
		// NOTE: placing this after setFocus() inexplicably solved a strange bug where if the first action
		//	during a session was to open a new taxon, the empty name prompt didn't work
		createNameListener(taxonBase.getName());
	}
	
	
	/**
	 * 
	 */
	private void showNameRelations() {
		TaxonNameBase name = getName();
		if (name == null) {
			return;
		}
		Set<NameRelationship> nameRelations = name.getNameRelations();
		if (nameRelations.size() == 0) {
			return;
		}
		for (NameRelationship nameRelation : nameRelations) {
			String typeLabel = null;
			TaxonNameBase relatedName = null; 
			
			boolean isZoological = name instanceof ZoologicalName ? true : false; 
			
			if (name.equals(nameRelation.getFromName())) {
				typeLabel = GlobalController.getNameRelationTypeLabel( 
					nameRelation.getType(), isZoological);
				relatedName = nameRelation.getToName();
			} else {
				typeLabel = GlobalController.getNameRelationInverseTypeLabel( 
					nameRelation.getType(), isZoological);
				relatedName = nameRelation.getFromName();
			}
			
			setNonEditableInfo(typeLabel + " " + CdmUtil.getDisplayName(relatedName));
		}
	}

	/**
	 * Listens for changes to this name's <code>fullTitleCache</code>.
	 * 
	 * @param data
	 */
	private void createNameListener(final TaxonNameBase name) {
		
		if (name == null) {
			return;
		}
		
		final PropertyChangeListener listener = new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent evt) {
				
				if (isParsing) {
					return;
				}
				
				if (EditorController.isSaving()) {
					return;
				}
				
				deactivateParser();
				textViewer.setText(name.getFullTitleCache());
				activateParser();
			}
		};
		
		// TODO clean this part up
		name.addPropertyChangeListener("fullTitleCache", listener);
		name.addPropertyChangeListener("nomenclaturalMicroReference", listener);

//		name.addPropertyChangeListener(ITaxEditorConstants.REFRESH_NAMEVIEWER, listener);
		
		ReferenceBase reference = name.getNomenclaturalReference();
		if (reference != null) {
			reference.addPropertyChangeListener("titleCache", listener);
		}
	}
	
	protected void initTextViewer(TaxonBase taxonBase) {
		
		showNameRelations();
		
		String text = CdmUtil.getDisplayNameWithRef(taxonBase);
		
		if (text.length() == 0) {
			initEmptyText();
		} else {
			textViewer.getTextWidget().setText(text);
			textViewer.setCursorToEOL();
		}
		calculateErrors();
		activateParser();
	}
	
	protected void parse(String text) {
		
		// Either composite is not yet fully built, or
		//	the property sheet is writing to it
		if (!isUseParser) {
			return;
		}
		
		// Let others know the parser is active
		isParsing = true;
						
		// Parse the name and paint the text field w any errors
		if (getName() != null) {
			CdmParserController.parseFullReference((NonViralName) getName(), text);
		}

		// Any entry of text means the taxon has been changed
		setDirty(true);
		
		// The parser is no longer active
		isParsing = false;
		
		// Manually refresh the property sheet to reflect changes
		setSelection();
		
		// Show any errors in the name viewer
		calculateErrors();
	}
	
	private void activateParser() {
		isUseParser = true;
	}
	
	private void deactivateParser() {
		isUseParser = false;
	}
	
	protected abstract TaxonNameBase getName();
	
	protected void calculateErrors() {
		textViewer.clearErrors();
		textViewer.setShowParsingError(getName());
	}

	protected void handleSplitText(String text) {
		// Create a synonym in a new homotypic group using text as name
		IUndoContext undoContext = EditorController.getUndoContext(taxon);
		IUndoableOperation operation = new CreateSynonymInNewGroupOperation
				("new heterotypic synonym", undoContext, taxon, text); //$NON-NLS-1$
		GlobalController.executeOperation(operation);
	}
	
	protected IPropertySource getPropertySourceByName(TaxonNameBase name) {
		if (name == null) {
			return null;
		}
		
		if (name.getClass() == BotanicalName.class) {
			return new BotanicalNamePropertySource((BotanicalName) name);
		}
		if (name.getClass() == ZoologicalName.class) {
			return new ZoologicalNamePropertySource((ZoologicalName) name);
		}
		if (name instanceof NonViralName) {
			return new NonViralNamePropertySource((NonViralName) name);
		}
		
		return null;
	}
}