/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.propertysheet.description;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertySource;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;

import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.model.common.Language;
import eu.etaxonomy.cdm.model.description.CommonTaxonName;
import eu.etaxonomy.cdm.model.description.DescriptionElementBase;
import eu.etaxonomy.cdm.model.description.Distribution;
import eu.etaxonomy.cdm.model.description.Feature;
import eu.etaxonomy.cdm.model.description.IndividualsAssociation;
import eu.etaxonomy.cdm.model.description.QuantitativeData;
import eu.etaxonomy.cdm.model.description.TaxonInteraction;
import eu.etaxonomy.cdm.model.description.TextData;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.editor.reference.IReferenceSearch;
import eu.etaxonomy.taxeditor.model.DescriptionUtil;
import eu.etaxonomy.taxeditor.model.ReferenceUtil;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferencePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferenceSearchDescriptor;

/**
 * @author p.ciardelli
 * @created 29.08.2008
 * @version 1.0
 */
public class DescriptionElementPropertySource implements IPropertySource {
	private static final Logger logger = Logger
			.getLogger(DescriptionElementPropertySource.class);
	
	/**
	 * The DescriptionElement whose properties are being displayed
	 */
	private DescriptionElementBase descriptionElement;
	
	/**
	 * If this is a property with a parent, the parent's property ID
	 */
	private String parentid;
	
	/**
	 * Property unique keys
	 */
	public static final String P_ID_ELEMENT_CLASS = "element_class";
	public static final String P_ID_ELEMENT_TEXT = "text";
	public static final String P_ID_FEATURE = "feature";
	public static final String P_ID_BIBREF = "bibref";
	public static final String P_ID_MICROREF = "microref";
	
	/**
	 * Property display keys
	 *  Note: for an explanation of the sorting prefixes ("04:"),
	 *  @see eu.etaxonomy.taxeditor.propertysheet.EditorPropertySheetEntry
	 */
	public static final String P_ELEMENT_CLASS = "01:Save element as";
	public static final String P_ELEMENT_TEXT = "00:Element";
	public static final String P_FEATURE = "01:Feature";
	public static final String P_BIBREF = "02:Bibliographic Reference";
	public static final String P_MICROREF = "03:Reference Detail";
	
	protected static final String[] TOP_LEVEL_PROPERTIES = 
			new String[] {P_ID_ELEMENT_TEXT, P_ID_FEATURE, P_ID_BIBREF, P_ID_MICROREF};
	
	private String[] P_FEATURE_CLASS_NAME_MENU;
	private Class[] P_FEATURE_CLASSES;
	
	/**
	 * Constructor for top level property fields. All fields that are not subfields
	 *  should be listed here.
	 * @param name
	 */
	public DescriptionElementPropertySource(DescriptionElementBase descriptionElement) {
		this(descriptionElement, null, TOP_LEVEL_PROPERTIES);
	}
	
	public DescriptionElementPropertySource(DescriptionElementBase descriptionElement, String parentid, 
			String[] keys) {
		this.descriptionElement = descriptionElement;
		this.parentid = parentid;
		for (String key : keys) {
			addDescriptor(key);
		}
	}

	protected Vector<PropertyDescriptor> descriptors = new Vector<PropertyDescriptor>();
	/**
	 * Add descriptor for a given property.
	 * 
	 * Notes on Descriptor:
	 * 
	 *  PropertyDescriptor - uneditable cell value
	 *  ComboBoxPropertyDescriptor - dropdown list, property supplied must be integer-based
	 *  TextPropertyDescriptor - editable cell value
	 *  
	 *  If any descriptor calls setCategory, all descriptors w/out a category are put in 
	 *  category "Misc".
	 *  
	 *  descriptor.setFilterFlags (new String[] { IPropertySheetEntry.FILTER_ID_EXPERT }) -
	 *   this descriptor shown under advanced properties
	 *  
	 * @param id
	 */
	protected void addDescriptor(String id) {
		if (id.equals(P_ID_ELEMENT_CLASS)) {
			
			Feature feature = descriptionElement.getFeature();
			Map<Class, String> supportedClasses = new HashMap<Class, String>();
			
			// Populate feature class arrays with classes supported by this feature
			if (feature.supportsCommonTaxonName()) {
				supportedClasses.put(CommonTaxonName.class, "Common name");
			}
			if (feature.supportsDistribution()) {
				supportedClasses.put(Distribution.class, "Distribution");				
			}
			if (feature.supportsIndividualAssociation()) {
				supportedClasses.put(IndividualsAssociation.class, "Individual association");				
			}
			if (feature.supportsQuantitativeData()) {
				supportedClasses.put(QuantitativeData.class, "Quantitative data");				
			}
			if (feature.supportsTaxonInteraction()) {
				supportedClasses.put(TaxonInteraction.class, "Taxon interaction");				
			}
			if (feature.supportsTextData()) {
				supportedClasses.put(TextData.class, "Text");				
			} 
			
			P_FEATURE_CLASS_NAME_MENU = supportedClasses.values().toArray((new String[supportedClasses.size()]));
			P_FEATURE_CLASSES = supportedClasses.keySet().toArray(new Class[supportedClasses.size()]);
			
			descriptors.addElement(				
					new ComboBoxPropertyDescriptor(P_ID_ELEMENT_CLASS, P_ELEMENT_CLASS, P_FEATURE_CLASS_NAME_MENU)
			);
		}
		
		if (id.equals(P_ID_FEATURE)) {
			descriptors.addElement(new PropertyDescriptor(P_ID_FEATURE, P_FEATURE));
		}
		
		if (id.equals(P_ID_BIBREF)) {			
			descriptors.addElement(new PropertyDescriptor(P_ID_BIBREF, P_BIBREF));
		}
		
		if (id.equals(P_ID_ELEMENT_TEXT)) {
			descriptors.addElement(new TextPropertyDescriptor(P_ID_ELEMENT_TEXT, P_ELEMENT_TEXT));
		}
		
		if (id.equals(P_ID_MICROREF)) {
			descriptors.addElement(new TextPropertyDescriptor(P_ID_MICROREF, P_MICROREF));
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getEditableValue()
	 */
	public Object getEditableValue() {
		return this;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyDescriptors()
	 */
	public IPropertyDescriptor[] getPropertyDescriptors() {
		return (IPropertyDescriptor[]) descriptors.toArray(
                new IPropertyDescriptor[descriptors.size()]);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyValue(java.lang.Object)
	 */
	public Object getPropertyValue(Object id) {
		if (id.equals(P_ID_ELEMENT_CLASS)) {
			if (descriptionElement instanceof TextData) {
				return 0;
			}
			if (descriptionElement instanceof CommonTaxonName) {
				return 1;
			}
			if (descriptionElement instanceof Distribution) {
				return 2;
			}
		}
		
		if (id.equals(P_ID_FEATURE)) {
			Feature feature = descriptionElement.getFeature();
			if (feature != null) {
				return feature.getLabel();
			} else {
				return "";
			}
		}

		if (id.equals(P_ID_BIBREF)) {
			ReferenceBase citation = descriptionElement.getCitation();
			ReferencePropertySource bibRefPropertySource = new ReferencePropertySource(citation);
			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					descriptionElement.setCitation((ReferenceBase) evt.getNewValue());
				}
			});
			return bibRefPropertySource;
		}
		
		if (id.equals(P_ID_ELEMENT_TEXT)) {
			return DescriptionUtil.getCache(descriptionElement);	
		}

		if (id.equals(P_ID_MICROREF)) {
			return CdmUtils.Nz(descriptionElement.getCitationMicroReference());
		}
		
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#isPropertySet(java.lang.Object)
	 */
	public boolean isPropertySet(Object id) {
		return false;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#resetPropertyValue(java.lang.Object)
	 */
	public void resetPropertyValue(Object id) {}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object, java.lang.Object)
	 */
	public void setPropertyValue(Object id, Object value) {
		if (id.equals(P_ID_ELEMENT_CLASS)) {
			// Hmmm ...
		}
		
		if (id.equals(P_ID_BIBREF)) {
			
			if (value instanceof ReferenceBase) {
				descriptionElement.setCitation((ReferenceBase) value);
			}
			
			ReferenceBase bibRef = 
					ReferenceUtil.getOrCreateCitation(descriptionElement);
			
			// "value" is the ReferencePropertySource, so we get the TitleCache from its
			// toString() method
			bibRef.setTitleCache(value.toString());
		}
		
		if (id.equals(P_ID_ELEMENT_TEXT)) {
			DescriptionUtil.setCache(descriptionElement, (String) value);
		}
		
		if (id.equals(P_ID_MICROREF)) {
			descriptionElement.setCitationMicroReference((String) value);
		}
		
		descriptionElement.firePropertyChange(ITaxEditorConstants.PROPERTY_SHEET_CHANGE, null, null);
	}
}
