/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */
package eu.etaxonomy.taxeditor.propertysheet.type;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertySource;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;

import eu.etaxonomy.cdm.model.name.NameTypeDesignation;
import eu.etaxonomy.cdm.model.name.SpecimenTypeDesignation;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.name.TypeDesignationBase;
import eu.etaxonomy.cdm.model.reference.Generic;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.taxeditor.propertysheet.CollectionPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferencePropertySource;

public class TypeCollectionPropertySource extends CollectionPropertySource {
	private static final Logger logger = Logger
			.getLogger(TypeCollectionPropertySource.class);
	
	private TaxonNameBase name;

	public TypeCollectionPropertySource(TaxonNameBase name, Collection collection) {
		super(collection);
		this.name = name;
	}
	
	protected void addDescriptor(Object item) {
		String itemDisplayName = getItemDisplayName(item);
		
		if (item instanceof NameTypeDesignation) {
			descriptors.addElement(
					new PropertyDescriptor(item, itemDisplayName));
		} else {
			
			/* Two issues here, caused by making a ProperySheet field editable while at the same
			 *  time returning an IPropertySource with which its child fields are built.
			 *  
			 * First, TextPropertyDescriptor expects its value to be set w/ a String and 
			 *  throws an Assert exception when this is not the case, i.e. when it is passed an
			 *  IPropertySource; hence, the override of doSetValue().
			 *  
			 * Second, the super implementation of doGetValue() was causing the equality test in
			 *  PropertySheetEntry:140 to return false: it was comparing the IPropertySource with 
			 *  the CellEditor's text String, thereby triggering an infinite loop where the cell's
			 *  previous state would never equals the cell's current state. Hence, the 
			 *  IPropertySource is now saved in the field editorValue for comparison.
			 */
			descriptors.addElement(
					new TextPropertyDescriptor(item, itemDisplayName));
//					new TextPropertyDescriptor(item, itemDisplayName) {
//					    public CellEditor createPropertyEditor(Composite parent) {
//					        CellEditor editor = new TextCellEditor(parent) {
//					        	Object editorValue;
//					            protected void doSetValue(Object value) {
//					            	if (value instanceof String) {
//					            		super.doSetValue(value);
//					            	} else {
//					            		editorValue = value;
//					            		super.doSetValue(value.toString());
//					            	}				            		
//					            }
////					            protected Object doGetValue() {
////					            	return super.doGetValue();
////					            }
//					        };
//					        if (getValidator() != null) {
//								editor.setValidator(getValidator());
//							}
//					        return editor;
//					    }
//					});
		}
	}
	
	@Override
	protected String getItemDisplayName(Object item) {
		String str = "";
		
		if (item instanceof SpecimenTypeDesignation) {
			
			SpecimenTypeDesignation typeDesignation = (SpecimenTypeDesignation) item;
			
			if (typeDesignation.getTypeStatus() != null) {
				str = typeDesignation.getTypeStatus().getLabel();
			} else {
				str = "UNDEFINED";
			}
		}

		if (item instanceof NameTypeDesignation) {
			
			NameTypeDesignation typeDesignation = (NameTypeDesignation) item;
			
			if (typeDesignation.isLectoType()) {
				str = "Lectotype";
			} else {
				str = "Type";
			}
		}
		
		return str;
	}

	@Override
	public Object getPropertyValue(Object id) {
		
		if (id instanceof TypeDesignationBase) {
			return new TypeDesignationPropertySource((TypeDesignationBase) id);
		}
		return null;
	}	
	
	@Override
	public void setPropertyValue(Object id, Object value) {
				
		if (id instanceof SpecimenTypeDesignation && value instanceof String) {
			SpecimenTypeDesignation typeDesignation = (SpecimenTypeDesignation) id;
			
			if (typeDesignation.getTypeSpecimen() != null) {
				typeDesignation.getTypeSpecimen().setTitleCache((String) value);
			} else {
				logger.warn("'typeDesignation.getTypeSpecimen()' returned null.");
			}
		}
	}

	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.propertysheet.CollectionPropertySource#toString()
	 */
	@Override
	public String toString() {
		return "";
	}
	
	class TypeDesignationPropertySource implements IPropertySource {

		private TypeDesignationBase typeDesignation;

		TypeDesignationPropertySource(TypeDesignationBase typeDesignation) {
			this.typeDesignation = typeDesignation;
		}
		
		public Object getEditableValue() {
			if (typeDesignation instanceof SpecimenTypeDesignation) {
				return ((SpecimenTypeDesignation) typeDesignation).getTypeSpecimen().getTitleCache();
			}
			
			if (typeDesignation instanceof NameTypeDesignation) {
				if (((NameTypeDesignation) typeDesignation).getTypeName() != null) {
					return ((NameTypeDesignation) typeDesignation).getTypeName().getTitleCache();
				}
			}
			
			return null;
		}

		public IPropertyDescriptor[] getPropertyDescriptors() {
			return new IPropertyDescriptor[]{
					new PropertyDescriptor(
							typeDesignation, "Citation")
			};
		}

		public Object getPropertyValue(Object id) {
			ReferenceBase reference = typeDesignation.getCitation();
			if (reference == null) {
				reference = Generic.NewInstance();
			}
			ReferencePropertySource referencePropertySource = new ReferencePropertySource(reference);
			referencePropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getNewValue() instanceof ReferenceBase) {	
						typeDesignation.setCitation((ReferenceBase) evt.getNewValue());
					}
				}
			});
			return referencePropertySource;
		}

		public boolean isPropertySet(Object id) {
			return false;
		}

		public void resetPropertyValue(Object id) {}

		public void setPropertyValue(Object id, Object value) {
		}
		
//		public String toString() {
//			if (typeDesignation instanceof SpecimenTypeDesignation) {
//				return ((SpecimenTypeDesignation) typeDesignation).getTypeSpecimen().getTitleCache();
//			}
//			
//			if (typeDesignation instanceof NameTypeDesignation) {
//				if (((NameTypeDesignation) typeDesignation).getTypeName() != null) {
//					return ((NameTypeDesignation) typeDesignation).getTypeName().getTitleCache();
//				}
//			}
//			
//			return null;
//		}
		
		public boolean equals(Object object) {
			if (object == null) {
				return false;
			}
			if (toString() == null) {
				return false;
			}
			if (toString().equals(object.toString())) {
				return true;
			} else {
				return false;
			}
		}
	}
}