// $Id$
/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/
package eu.etaxonomy.taxeditor.annotatedlineeditor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.ITypedRegion;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TypedRegion;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.jface.text.rules.IPartitionTokenScanner;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.text.rules.RuleBasedPartitionScanner;
import org.eclipse.jface.text.rules.Token;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.texteditor.AbstractDocumentProvider;



/**
 * Using an <code>IEditorPart</code>, creates a document where each line is associated
 * with an entity. Mapping between document positions and entities is stored in
 * <code>LineAnnotation</code>s in a <code>LineAnnotationModdel</code>.  
 * <p>
 * Requires:
 * <ul>
 * <li>an <code>IEntityCreator</code> to create entities for new lines;
 * <li>an <code>IEntityPersistenceService</code> for interacting with the persistence layer; and
 * <li>an <code>ILineDisplayStrategy</code> for various visual manifestations of the domain object. 
 * </ul> 
 * 
 * @author p.ciardelli
 * @created 25.06.2009
 * @version 1.0
 */
public class AnnotatedLineDocumentProvider extends AbstractDocumentProvider {

	private static final Logger logger = Logger
			.getLogger(AnnotatedLineDocumentProvider.class);
	
	private Map<IEditorInput, IDocument> documents = 
			new HashMap<IEditorInput, IDocument>();

	private Map<IEditorInput, IAnnotationModel> annotationModels = 
		new HashMap<IEditorInput, IAnnotationModel>();

	private Map<IEditorInput, IEntityCreator<?>> entityCreators = 
		new HashMap<IEditorInput, IEntityCreator<?>>();

	private Map<IEditorInput, IEntityPersistenceService> persistenceServices = 
		new HashMap<IEditorInput, IEntityPersistenceService>();

	private Map<IEditorInput, ILineDisplayStrategy> lineDisplayStrategies = 
		new HashMap<IEditorInput, ILineDisplayStrategy>();
	
	@Override
	public IAnnotationModel getAnnotationModel(Object element) {
		if (element instanceof IEditorInput) {
			
			IEditorInput input = (IEditorInput) element;
			IAnnotationModel model = annotationModels.get((IEditorInput) element);
			
			// Create model as necessary
			if (model == null) {
				model = new LineAnnotationModel(getLineDisplayStrategy(element));
				((LineAnnotationModel) model).setEntityCreator(getEntityCreator(element));
				annotationModels.put(input, model);
			}
			return model;
		}
		
		return null;
	}
	
	@Override
	protected IAnnotationModel createAnnotationModel(Object element)
			throws CoreException {
		return getAnnotationModel(element);
	}

	@Override
	public IDocument getDocument(Object element) {
		if (element instanceof IEditorInput) {
			return documents.get((IEditorInput) element);
		}
		return null;
	}
	
	@Override
	protected IDocument createDocument(Object element) throws CoreException {
		
		if (element instanceof IEditorInput) {
			IEditorInput input = (IEditorInput) element;
			Document document = new Document("");
			IAnnotationModel model = getAnnotationModel(element);
			documents.put(input, document);
					
			for (Object entity : getEntityList(element)) {
				try {
					createAnnotatedLine(input, entity);
				} catch (BadLocationException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}	
			
			return document;
		}
		return null;
	}

	/**
	 * Creates an annotated line at the end of the document associated with the element
	 * @return 
	 * @throws BadLocationException 
	 */
	protected LineAnnotation createAnnotatedLine(Object element, Object entity) throws BadLocationException {
		
		Document document = (Document) getDocument(element);
		ILineDisplayStrategy lineDisplay = getLineDisplayStrategy(element);
		
		LineAnnotation annotation = new LineAnnotation(entity, lineDisplay);
				
		// Is document zero length, or is last char in document line delimiter?
		int docLength = document.getLength();
		boolean useDelimiter = false;
		if (docLength > 0) {
			if (docLength > 1 && !document.get(docLength - 2, 2).equals(document.getDefaultLineDelimiter())) {
				useDelimiter = true;
			}
		}
		if (useDelimiter) {
			document.replace(docLength, 0, document.getDefaultLineDelimiter());
		}
		
		String text;
		if (lineDisplay.isEntityCacheEmpty(entity)) {
			text = lineDisplay.getEmptyCacheMessage(entity);
		} else {
			text = lineDisplay.getText(entity);
		}
		text += lineDisplay.getSupplementalText(entity);
		docLength = document.getLength();
		document.replace(docLength, 0, text);
		
		Position position = new Position(docLength, text.length());
		
		IAnnotationModel model = getAnnotationModel(element);
		model.addAnnotation(annotation, position);
					
		logger.info("Annotation added " + entity);
							
		lineDisplay.addDisplayListener(entity, 
						new EntityListenerImpl((LineAnnotation) annotation, element, lineDisplay));
				
		return annotation;
	}
	
	public void updateLineFromAnnotation(Object element, LineAnnotation annotation) {
				
		IAnnotationModel model = getAnnotationModel(element);
		IDocument document = getDocument(element);
		ILineDisplayStrategy lineDisplay = getLineDisplayStrategy(element);
		
		Object entity = annotation.getEntity();
		String text = "";
		if (getLineDisplayStrategy(element).isEntityCacheEmpty(entity)) {
			text = lineDisplay.getEmptyCacheMessage(entity);
		} else {
			text = lineDisplay.getText(entity);
		}		
		text += lineDisplay.getSupplementalText(entity);
		
		try {
			if (model.getPosition(annotation) == null) {
				return;
			}
			int offset = model.getPosition(annotation).getOffset();			
			int line = document.getLineOfOffset(offset);
			int lineLength = document.getLineLength(document.getLineOfOffset(offset));
			if (document.getLineDelimiter(line) != null) {
				lineLength -= document.getLineDelimiter(line).length();
			}
			document.replace(offset, lineLength, text);
		} catch (BadLocationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	class EntityListenerImpl extends EntityListener {

		private LineAnnotation annotation;
		private ILineDisplayStrategy lineDisplay;
		private Object element;

		EntityListenerImpl(LineAnnotation annotation, Object element, ILineDisplayStrategy lineDisplay) {
			this.annotation = annotation;
			this.element = element;
			this.lineDisplay = lineDisplay;
		}
		/* (non-Javadoc)
		 * @see eu.etaxonomy.taxeditor.annotatedlineeditor.EntityDisplayListener#updateDisplay()
		 */
		@Override
		protected void updateEntity() {
			annotation.setDirty(true);
			updateLineFromAnnotation(element, annotation);
		}
	}
	
	/**
	 * @param element 
	 * @return
	 */
	private List<?> getEntityList(Object element) {
		if (element instanceof IEditorInput) {
			return getPersistenceService(element).getEntityList(element);
		}
		return null;
	}

	@Override
	protected void doSaveDocument(IProgressMonitor monitor, Object element,
			IDocument document, boolean overwrite) throws CoreException {
		if (element instanceof IEditorInput) {
			
			IEntityPersistenceService persistenceService = getPersistenceService(element);
			
			// Get new containers from annotation model
			LineAnnotationModel model = (LineAnnotationModel) getAnnotationModel(element);
			Iterator iterator = model.getAnnotationIterator();
			while (iterator.hasNext()) {
				Annotation annotation = (Annotation) iterator.next();
				if (annotation instanceof IEntityContainer<?>) {
					IEntityContainer<?> container = (IEntityContainer<?>) annotation;
					if (container.isMarkedAsNew() || container.isDirty()) {
						persistenceService.save(container.getEntity(), element); // save
						container.setDirty(false);
						container.markAsNew(false);
					}
				}
			}
			for (LineAnnotation annotation : model.getDeletedAnnotations()) {
				if (annotation.isMarkedAsNew()) {
					continue;
				}
				if (annotation.isMarkedAsMerged()) {
					persistenceService.merge(annotation.getEntity(), annotation.getMergeTarget(), element); //  merge					
				} else {
					// TODO clarify w AM whether this needs to be executed on merged objects
					persistenceService.delete(annotation.getEntity(), element); // delete
				}
			}
			model.clearDeletedAnnotations();
		}
	}

	@Override
	protected IRunnableContext getOperationRunner(IProgressMonitor monitor) {
		return null;
	}
	
	@Override
	public boolean isModifiable(Object element) {
		return true;
	}
	
	@Override
	public boolean isReadOnly(Object element) {
		// enables copy & paste
		return false;
	}

	/**
	 * @param entityCreator
	 * @param element
	 */
	public void setEntityCreator(IEntityCreator<?> entityCreator, Object element) {
		if (element instanceof IEditorInput) {
			entityCreators.put((IEditorInput) element, entityCreator);
		}
	}
	
	/**
	 * @param element
	 * @return
	 */
	public IEntityCreator<?> getEntityCreator(Object element) {
		if (element instanceof IEditorInput) {
			IEntityCreator<?> entityCreator = entityCreators.get(element);
			Assert.isNotNull(entityCreator, "No IEntityCreator set for this element.");
			return entityCreator;
		}
		return null;
	}

	/**
	 * @param persistenceService
	 * @param input
	 */
	public void setPersistenceService(
			IEntityPersistenceService persistenceService, Object element) {
		if (element instanceof IEditorInput) {
			persistenceServices.put((IEditorInput) element, persistenceService);
		}
	}
	
	public IEntityPersistenceService getPersistenceService(Object element) {
		if (element instanceof IEditorInput) {
			IEntityPersistenceService persistenceService = persistenceServices.get((IEditorInput) element);
			Assert.isNotNull(persistenceService, "No IEntityPersistenceService set for this element.");
			return persistenceService;
		}
		return null;
	}

	/**
	 * @param lineDisplayStrategy
	 * @param input
	 */
	public void setLineDisplayStrategy(
			ILineDisplayStrategy lineDisplayStrategy, Object element) {
		if (element instanceof IEditorInput) {
			lineDisplayStrategies.put((IEditorInput) element, lineDisplayStrategy);
		}	
	}
	
	/**
	 * @param element
	 * @return
	 */
	protected ILineDisplayStrategy getLineDisplayStrategy(Object element) {
		if (element instanceof IEditorInput) {
			ILineDisplayStrategy lineDisplayStrategy = lineDisplayStrategies.get((IEditorInput) element);
			Assert.isNotNull(lineDisplayStrategy, "No ILineDisplayStrategy set for this element.");
			return lineDisplayStrategy;
		}
		return null;
	}
	
	/**
	 * @param editorInput
	 * @param lineno
	 */
	public void removeAnnotatedLine(Object element, int lineno) {
		Document document = (Document) getDocument(element);
		LineAnnotationModel model = (LineAnnotationModel) getAnnotationModel(element);
		LineAnnotation annotation = (LineAnnotation) model.getAnnotationAtLine(lineno, document);
		removeAnnotatedLine(element, annotation);
	}
	
	public void removeAnnotatedLine(Object element, LineAnnotation annotation) {
		if (annotation != null) {
			Document document = (Document) getDocument(element);
			LineAnnotationModel model = (LineAnnotationModel) getAnnotationModel(element);
			
			Position position = model.getPosition(annotation);
			int offset = position.getOffset();
			int length = position.getLength();
			
			Object entity = annotation.getEntity();
			annotation.markAsDeleted();
			model.removeAnnotation(annotation);
			
			// Immediately followed by a delimiter?
			int annotationEnd = offset + length;
			try {
				if (document.getLength() > annotationEnd + 1 && document.get(annotationEnd, 2).equals(document.getDefaultLineDelimiter())) {
					length += 2;
				}
			} catch (BadLocationException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			
			try {
				document.replace(offset, length, "");
			} catch (BadLocationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}		
	}
}