/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.editor.name;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.commands.operations.IUndoableOperation;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IStatusLineManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.dnd.DropTargetAdapter;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.forms.ManagedForm;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.TableWrapLayout;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.views.properties.IPropertySheetEntry;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.IPropertySource;
import org.eclipse.ui.views.properties.PropertySheetPage;
import org.eclipse.ui.views.properties.PropertySheetSorter;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;

import eu.etaxonomy.cdm.api.conversation.ConversationHolder;
import eu.etaxonomy.cdm.api.conversation.IConversationEnabled;
import eu.etaxonomy.cdm.hibernate.HibernateProxyHelper;
import eu.etaxonomy.cdm.model.common.CdmBase;
import eu.etaxonomy.cdm.model.name.HomotypicalGroup;
import eu.etaxonomy.cdm.model.taxon.Synonym;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonBase;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationship;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationshipType;
import eu.etaxonomy.cdm.persistence.hibernate.CdmDataChangeMap;
import eu.etaxonomy.taxeditor.editor.CompositeBorderDecorator;
import eu.etaxonomy.taxeditor.editor.EditorUtil;
import eu.etaxonomy.taxeditor.editor.GroupComposite;
import eu.etaxonomy.taxeditor.editor.GroupedComposite;
import eu.etaxonomy.taxeditor.editor.IHasPropertySource;
import eu.etaxonomy.taxeditor.editor.IMultiPageTaxonEditorPage;
import eu.etaxonomy.taxeditor.editor.MultiPageTaxonEditor;
import eu.etaxonomy.taxeditor.editor.SimpleSelectionProvider;
import eu.etaxonomy.taxeditor.editor.TaxonEditorInput;
import eu.etaxonomy.taxeditor.editor.WidgetTransfer;
import eu.etaxonomy.taxeditor.model.Resources;
import eu.etaxonomy.taxeditor.operations.ChangeConceptToSynonymOperation;
import eu.etaxonomy.taxeditor.operations.ChangeHomotypicGroupOperation;
import eu.etaxonomy.taxeditor.operations.IPostOperationEnabled;
import eu.etaxonomy.taxeditor.propertysheet.EditorPropertySheetEntry;
import eu.etaxonomy.taxeditor.propertysheet.PropertySheetUtil;

/**
 * @author p.ciardelli
 * @created 15.05.2008
 * @version 1.0
 */
public class TaxonNameEditor extends EditorPart implements IMultiPageTaxonEditorPage, IAdaptable, IConversationEnabled, IPostOperationEnabled, ITabbedPropertySheetPageContributor 
		,INameEditorCompositeRepository {
	
	private static final Logger logger = Logger.getLogger(TaxonNameEditor.class);
	
	private Taxon taxon;
	
	protected ManagedForm managedForm;
	protected ScrolledForm scrolledForm; 
	protected Composite parent;
	protected ISelectionProvider provider;
	
	protected MultiPageTaxonEditor editor;
	
	protected IHasPropertySource selectedObject;

	protected Composite partComposite;
	protected GroupedComposite firstGroupedComposite;

	protected Object selectedData;
	
	protected ConversationHolder conversation;

	private MenuManager menuManager;

	private Menu menu;

	/**
	 * The object that was affected by last operation.
	 */
	private CdmBase objectAffectedByLastOperation;

	private boolean isRedrawing;
	
	private static final String ID = "eu.etaxonomy.taxeditor.taxonNameEditor";

	private Map<HomotypicalGroup, HomotypicalGroupComposite> homotypicGroupComposites;

	private MisappliedGroupComposite misappliedGroupComposite;

	private ConceptGroupComposite conceptGroupComposite;
	
	
	
	public TaxonNameEditor(MultiPageTaxonEditor editor){
		this.editor = editor;
		conversation = editor.getConversationHolder();
	}
	
	/**
	 * Shared listener that sets dirty state to true 
	 *  when any registered property changes 
	 */
	private PropertyChangeListener taxonChangeListener = new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent event) {
			if (event.getPropertyName().equals(Resources.PROPERTY_SHEET_CHANGE)) {
				firePropertyChange(PROP_DIRTY);
			}
		}
	};

	private DropTarget target;
	
	public IUndoContext getUndoContext() {
		return editor.getUndoContext();
	}
	
	/**
	 * If there is a default property sheet object with a corresponding property source class, 
	 * display it in the property sheet. Otherwise, empty the property sheet with an empty
	 * <code>StructuredSelection</code>.
	 * 
	 * @param selectedObject
	 */
	protected void setSelection(IHasPropertySource selectedObject) {
				
		// Unpaint last selection - last selection will only be unpainted
		//	when something else on this page is selected
		if (this.selectedObject instanceof GroupedComposite) {
			GroupedComposite composite = ((GroupedComposite) this.selectedObject);
			composite.unpaintBorder();
			composite.setBackground(Resources.getColor(Resources.COLOR_COMPOSITE_BACKGROUND));
		}
		
		// Set the selection to this editor's selected object
		this.selectedObject = selectedObject;
		
		//  set background on selection TODO wanted to move this to setFocus() on the selectedObject but that doesn't work (infinite loop)
//		((GroupedComposite) selectedObject).setBackground(Resources.getColor(Resources.COLOR_COMPOSITE_SELECTED));
		
		
		if (selectedObject instanceof Composite) {
			((GroupedComposite) selectedObject).setBackground(Resources.getColor(Resources.COLOR_COMPOSITE_SELECTED));
			this.selectedData = ((Composite) selectedObject).getData();
		}
		
		// Get the selection's property source, pass it to the selection provider
		IPropertySource propertySource = null;
		if (selectedObject != null) {
			propertySource = selectedObject.getPropertySource();
		}
		if (propertySource == null) {
			provider.setSelection(new StructuredSelection());
		} else {
			provider.setSelection(new StructuredSelection(propertySource));
		}
		
//		// FIXME TabbedPropertySheet stuff 
//		provider.setSelection(new StructuredSelection(this.selectedObject));
		
	}


	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite composite) {

		this.partComposite = composite;
		
		createManagedForm(composite);
	}
	
	protected void createManagedForm(Composite composite) {
				
		managedForm = new ManagedForm(composite) {
			public void dirtyStateChanged() {
				firePropertyChange(PROP_DIRTY);
			}
			public boolean setInput(Object input) {
				if (input instanceof IHasPropertySource) {
					setSelection((IHasPropertySource)input);
				}
				return super.setInput(input); 
			} 
		};
		
		scrolledForm = managedForm.getForm();
		parent = scrolledForm.getBody();
		
		// register the context menu
		menuManager = new MenuManager();
		ISelectionProvider selectionProvider = getSite().getSelectionProvider();
		getSite().registerContextMenu(TaxonNameEditor.ID, menuManager, selectionProvider);
		
		menu = menuManager.createContextMenu(parent);
		scrolledForm.setMenu(menu);
				
		parent.setData(taxon);
		
		parent.setLayout(new TableWrapLayout());		
		parent.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		
		Taxon taxon = (Taxon) HibernateProxyHelper.deproxy(getTaxon());
		
		firstGroupedComposite = createAcceptedTaxon();
				
		// Draw homotypic group synonyms
		HomotypicalGroup homotypicGroupOfAcceptedTaxon = (HomotypicalGroup) HibernateProxyHelper.deproxy(taxon.getHomotypicGroup());
		
		if (homotypicGroupOfAcceptedTaxon != null) {
			List<Synonym> homotypicSynonyms = homotypicGroupOfAcceptedTaxon.getSynonymsInGroup(taxon.getSec());
			for (Synonym synonym : homotypicSynonyms) {
				createSynonym(synonym);
			}
		}
		
		// Draw heterotypic synonym groups
		List<HomotypicalGroup> heterotypicSynonymGroups = taxon.getHeterotypicSynonymyGroups();
		for (HomotypicalGroup homotypicalGroup : heterotypicSynonymGroups) {
							
			List<Synonym> heterotypicSynonyms = homotypicalGroup.
					getSynonymsInGroup(taxon.getSec());
			for (Synonym synonym : heterotypicSynonyms) {
				 createSynonym(synonym);
			}
		}
		
		// Draw misapplied name elements
		for(Taxon misappliedName : taxon.getMisappliedNames()){
			createMisappliedName(misappliedName);
		}		

		// Draw concept relation elements
		Set<TaxonRelationship> taxonRelations = taxon.getTaxonRelations();
		for (TaxonRelationship relationship : taxonRelations) {
			
			if (relationship.getType().equals(TaxonRelationshipType.MISAPPLIED_NAME_FOR()) ||
					relationship.getType().equals(TaxonRelationshipType.TAXONOMICALLY_INCLUDED_IN())) {
				continue;
			}
							
			createConcept(relationship);
		}		
			
		createDragSupport();
		
		setFocus();

		// Redraw composite
		composite.layout();
	}

	@SuppressWarnings("unchecked")
	public Object getAdapter(Class type) {
	    if (type == IPropertySheetPage.class) {
	    	
	    	PropertySheetPage page = new EditorPropertySheetPage();
	        
	        PropertySheetUtil.setPropertySheetPage(page);
	        
	        EditorPropertySheetEntry entry = new EditorPropertySheetEntry(taxon, this);
	        page.setRootEntry(entry);
	        page.refresh();
	        
	        return page;
	    	
//	        // FIXME TabbedPropertySheet stuff
//	    	EditorTabbedPropertySheetPage page = new EditorTabbedPropertySheetPage(this);
//	    	return page;	    	
	    }
	    
	    return super.getAdapter(type);
	}
	
	public Taxon getTaxon() {
		return taxon;
	}
	
	public void setDirty() {
		managedForm.dirtyStateChanged();
	}
	
	
	
	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	@Override
	public void setFocus(){
		Object selection;
		
		if(selectedData == null){
			selection = (objectAffectedByLastOperation == null) ? null : objectAffectedByLastOperation;
		}else{
			selection = (objectAffectedByLastOperation == null) ? selectedData : objectAffectedByLastOperation;
		}
				
		// make selection
		if (selection == null){
			GroupedComposite focusComposite = firstGroupedComposite;
			setSelection(focusComposite);
			focusComposite.drawBorder();
		} else {
			selectedData = selection;
			
			for (GroupedComposite composite : getGroupedComposites()) {
				if (selectedData.equals(composite.getData())) {
					setSelection((GroupedComposite) composite);
					((GroupedComposite) composite).drawBorder();
					((GroupedComposite) composite).setSelected();
					break;
				}					
			}
		}
		
		// reset
		objectAffectedByLastOperation = null;
	}
	
	/**
	 * This method returns all <code>GroupedComposite</code> controls currently attached to this editor.
	 * 
	 * @return a <code>List</code> of <code>GroupedComposite</code> controls.
	 */
	public List<GroupedComposite> getGroupedComposites(){
		return getGroupedCompositesRecursively(scrolledForm.getBody(), new ArrayList<GroupedComposite>());
	}

	
	/**
	 * Recursively traverse the composite hierarchy and collect all <code>GroupedComposite</code> controls.
	 * 
	 * @param composite the composite to start recursing from
	 * @param groupedComposites a <code>List</code> the found composites are stored in 
	 * @return a <code>List</code> containing all found <code>GroupedComposite</code> controls
	 */
	private List<GroupedComposite> getGroupedCompositesRecursively(Composite composite, List<GroupedComposite> groupedComposites){
		for(Control child : composite.getChildren()){
			if(child instanceof GroupedComposite){
				groupedComposites.add((GroupedComposite) child);
			}else if(child instanceof Composite){
				getGroupedCompositesRecursively((Composite) child, groupedComposites);
			}
		}
		return groupedComposites;
	}
	
	/**
	 * @return
	 */
	public List<GroupComposite> getGroupComposites() {
		return getGroupCompositesRecursively(scrolledForm.getBody(), new ArrayList<GroupComposite>());
	}
	
	/**
	 * Recursively traverse the composite hierarchy and collect all <code>GroupComposite</code> controls.
	 * 
	 * @param composite the composite to start recursing from
	 * @param groupComposites a <code>List</code> the found composites are stored in 
	 * @return a <code>List</code> containing all found <code>GroupComposite</code> controls
	 */
	private List<GroupComposite> getGroupCompositesRecursively(Composite composite, List<GroupComposite> groupComposites){
		for(Control child : composite.getChildren()){
			if(child instanceof GroupComposite){
				groupComposites.add((GroupComposite) child);
				getGroupCompositesRecursively((Composite) child, groupComposites);
			}
		}
		return groupComposites;
	}
	
	public void setInput(IEditorInput input){
		this.setInputWithNotify(input);		
	}
	
	/*
	 * (non-Javadoc)
	 * @see eu.etaxonomy.cdm.api.conversation.IConversationEnabled#getConversationHolder()
	 */
	public ConversationHolder getConversationHolder(){
		return conversation;
	}
	
	/*
	 * (non-Javadoc)
	 * @see eu.etaxonomy.cdm.persistence.hibernate.ICdmPostCrudObserver#update(eu.etaxonomy.cdm.persistence.hibernate.CdmCrudEvent)
	 */
	public void update(CdmDataChangeMap events) {
		//redraw();
	}
	
	/**
	 * Redraws this editor
	 * return true on success
	 */
	public boolean redraw(){
		return redraw(true);
	}
	
	/**
	 * Redraws the editor controls
	 * 
	 * @param focus if set to true the focus is set on the editor
	 * @return true on success
	 */
	public boolean redraw(boolean focus) {
		
//		for(NameComposite nameComposite : getNameComposites()){
//			nameComposite.persistName();
//		}
		
		isRedrawing = true;
		
		this.selectedObject = null;
		managedForm.getForm().dispose();
		this.resetGroupComposites();
		createManagedForm(partComposite);
		
		setFocus();
		
		isRedrawing = false;
		
		return true;
		
	}

	public boolean isRedrawing() {
		return isRedrawing;
	}
	
	public MultiPageTaxonEditor getMultiPageTaxonEditor() {
		return editor;
	}


	/*
	 * (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.store.operations.IPostOperationEnabled#postOperation()
	 */
	public boolean postOperation(CdmBase objectAffectedByOperation) {
		
		if (isRedrawing) {
			return false;
		}
		
		if (objectAffectedByOperation == null && selectedData instanceof CdmBase) {
			this.objectAffectedByLastOperation = (CdmBase) selectedData;	
		} else {
			this.objectAffectedByLastOperation = objectAffectedByOperation;
		}
		editor.setDirty();
		return redraw();	
	}
	

	/**
	 * @return the managedForm
	 */
	public ManagedForm getManagedForm() {
		return managedForm;
	}
	
	public Composite getTopLevelComposite() {
		return this.getManagedForm().getForm().getBody();
	}
	

	/**
	 * @return the menu
	 */
	public Menu getMenu() {
		return menu;
	}

	/**
	 * @return the firstGroupedComposite
	 */
	public GroupedComposite getFirstGroupedComposite() {
		return firstGroupedComposite;
	}

	class EditorPropertySheetPage extends PropertySheetPage {
		
		EditorPropertySheetPage() {
			super();
			
			// Override sorter to simply display names as first-in-first-out
			setSorter(new PropertySheetSorter() {
				public int compare(IPropertySheetEntry entryA, IPropertySheetEntry entryB) {
					return 0;
				}
				public int compareCategories(String categoryA, String categoryB) {
					return 0;
				}
				public void sort(IPropertySheetEntry[] entries) {
					// do nothing
				}
			});
		}
		
        public void makeContributions(IMenuManager menuManager,
                IToolBarManager toolBarManager, IStatusLineManager statusLineManager) {
        	super.makeContributions(menuManager, toolBarManager, statusLineManager);
        
        	// Remove "Show categories", "Show advanced properties", "Restore default value"
        	toolBarManager.removeAll();
        	menuManager.removeAll();
        }
        
        public Control getControl() {
        	Control control = super.getControl();
        	
        	// Save the property sheet tree for easy access as needed
        	if (!control.isDisposed()) {
        		if (control instanceof Tree) {
        			PropertySheetUtil.setPropertySheetTree((Tree) control);
        		}
        	}
        	return control;
        }
	}
	
	/**
	 * 
	 */
	public void navigateToNextMember() {
		try{
			List<GroupedComposite> composites = getGroupedCompositesRecursively(scrolledForm.getBody(), new ArrayList<GroupedComposite>());
			
			for(int i = 0; i < composites.size(); i++){
				if(selectedData == composites.get(i).getData()){
					selectedData = composites.get(i+1).getData();
					break;
				}
			}
			setFocus();
		}catch(IndexOutOfBoundsException e){}
	}
	
	/**
	 * 
	 */
	public void navigateToPreviousMember() {
		try{
			List<GroupedComposite> composites = getGroupedCompositesRecursively(scrolledForm.getBody(), new ArrayList<GroupedComposite>());
			
			for(int i = 0; i < composites.size(); i++){
				if(selectedData == composites.get(i).getData()){
					selectedData = composites.get(i-1).getData();
					break;
				}
			}
			setFocus();
		}catch(IndexOutOfBoundsException e){}
	}
	
	/**
	 * Used by ITabbedPropertySheet...
	 */
    public String getContributorId() {
        return TaxonNameEditor.ID;
    }


	/**
	 * @return true if there are empty names
	 */
	public boolean checkForEmptyNames() {
		for(NameComposite nameComposite : getNameComposites()){
			if(nameComposite.getName() == null || nameComposite.getName().getTitleCache().equals("")){
				MessageDialog.openWarning(EditorUtil.getShell(), "No Name Specified", "An attempt was made to save a taxon or synonym with " +
				"an empty name. Operation was cancelled.");
				return true;
			}
		}
		return false;
	}
    
	@Override
	public void doSave(IProgressMonitor monitor) {

		// check for empty names
		for(NameComposite nameComposite : getNameComposites()){
			nameComposite.persistName();	
		}
		
	}

	@Override
	public void doSaveAs() {}

	@Override
	public void init(IEditorSite site, IEditorInput input)
			throws PartInitException {
				
		if (!(input instanceof IEditorInput))
			throw new PartInitException(
				"Invalid Input: Must be IEditorInput");
		
		if (input.getAdapter(Taxon.class) != null) {
			taxon = (Taxon) input.getAdapter(Taxon.class);
		} else {
			throw new PartInitException(
				"Invalid Input: Taxon cannot be null");
		}

		setSite(site);
		setInput(input);
		
		provider = new SimpleSelectionProvider();
		getSite().setSelectionProvider(provider);
		
		
		logger.trace("New Editor instance created.");
		
		// Register listeners for any change in accepted name or set of relations
		
		// TODO what about property change listeners. is this still used?
		taxon.getName().addPropertyChangeListener(taxonChangeListener);
		taxon.addPropertyChangeListener(taxonChangeListener);
	}
		
	/**
	 * 
	 */
	private void createDragSupport() {
		// Listen for names being dragged outside of existing homotypic groups -
		//  user wants to create a new group
		Transfer[] types = new Transfer[] {WidgetTransfer.getInstance()};
		int operations = DND.DROP_MOVE | DND.DROP_COPY | DND.DROP_DEFAULT;
		if(target == null){
			target = new DropTarget(parent, operations);
			target.setTransfer(types);
			target.addDropListener(new NameEditorDropTargetListener());
		}		
	}

	public Composite getAcceptedName() {
		if(firstGroupedComposite instanceof AcceptedNameComposite){
			return firstGroupedComposite;
		}
		throw new RuntimeException("The accepted name should always reside in firstGroupedComposite");
	}

	/**
	 * @return
	 */
	private Map<HomotypicalGroup, HomotypicalGroupComposite> getHomotypicGroups() {
		if(homotypicGroupComposites == null){
			homotypicGroupComposites = new HashMap<HomotypicalGroup, HomotypicalGroupComposite>();
		}
		return homotypicGroupComposites;
	}
	
	public HomotypicalGroupComposite getHomotypicGroup(HomotypicalGroup group) {
		return getHomotypicGroups().get(group);
	}

	
	public Composite getMisappliedGroup() {
		if (misappliedGroupComposite == null) {
			misappliedGroupComposite = createMisappliedGroup();
		}
		return misappliedGroupComposite;
	}

	
	public Composite getMisappliedName(Taxon misappliedName) {
		for(Control child : getMisappliedGroup().getChildren()){
			if(child instanceof MisappliedNameComposite){
				return (MisappliedNameComposite) child;
			}
		}
		return null;
	}

	
	public Composite getSynonym(Synonym synonym) {
				
		GroupComposite groupComposite = getHomotypicGroup(synonym.getHomotypicGroup());
		
		for (Control child : groupComposite.getChildren()){
			if(child instanceof SynonymComposite 
					&& synonym.equals(((SynonymComposite) child).getSynonym())){
				return (SynonymComposite) child;
			}
		}
		
		return null;
	}

	
	public Composite getConceptGroup() {
		if (conceptGroupComposite == null) {
			conceptGroupComposite = createConceptGroup();
		}
		return conceptGroupComposite;
	}

	
	public Composite getConcept(Taxon relatedConcept) {
		for (Control child : getConceptGroup().getChildren()){
			if(child instanceof ConceptComposite 
					&& relatedConcept.equals 
					(((ConceptComposite) child).getRelatedTaxon())){
				return (ConceptComposite) child;
			}
		}

		return null;
	}
	
	/**
	 * @return a Set containing all composites that have been edited
	 */
	public Set<NameComposite<TaxonBase>> getDirtyNames(){
		Set<NameComposite<TaxonBase>> dirtyNames = new HashSet<NameComposite<TaxonBase>>();
		
		for(NameComposite composite : getNameComposites()){
			if(composite.isDirty()){
				dirtyNames.add(composite);
			}
		}
		
		return dirtyNames;
	}
	
	public Set<NameComposite<TaxonBase>> getNameComposites(){
		Set<NameComposite<TaxonBase>> nameComposites = new HashSet<NameComposite<TaxonBase>>();
		
		for(GroupedComposite composite : getGroupedComposites()){
			if(composite instanceof NameComposite){
				nameComposites.add((NameComposite) composite);
			}
		}
		
		return nameComposites;
	}
	
	/***********************************************************************/
	
	private GroupedComposite createAcceptedTaxon() {
		// Create a homotypic group composite for the accepted taxon
		HomotypicalGroup group = (HomotypicalGroup) HibernateProxyHelper.deproxy(getTaxon().getHomotypicGroup());
		Assert.isNotNull(group, "Taxon does not have a homotypic group");
		
		createHomotypicalGroup(group);
		
		// Create a name composite for the accepted taxon
		return new AcceptedNameComposite(this, getHomotypicGroup(group));
	}
	
	private HomotypicalGroupComposite createHomotypicalGroup(HomotypicalGroup group) {

		// Create the group composite
		HomotypicalGroupComposite groupComposite = new HomotypicalGroupComposite(this, getTopLevelComposite(), group);
		
		groupComposite.setData(FormToolkit.KEY_DRAW_BORDER, FormToolkit.TEXT_BORDER);
		getManagedForm().getToolkit().paintBordersFor(groupComposite.getParent());
		
		getHomotypicGroups().put(group, groupComposite);
		
		return groupComposite;
	}


	private Composite createSynonym(Synonym synonym) {		
		// Get the synonym's homotypical group
		HomotypicalGroup group = (HomotypicalGroup) HibernateProxyHelper.deproxy(synonym.getHomotypicGroup());
		Assert.isNotNull(group, "Synonym does not have a homotypic group");
		
		// If the group doesn't yet have a composite, create one and add it to the repository
		HomotypicalGroupComposite groupComposite = getHomotypicGroup(group);
		if (groupComposite == null) {
			groupComposite = createHomotypicalGroup(group);
		}
				
		// Create a synonym composite in the homotypical group
		Composite synonymComposite = new SynonymComposite(this, groupComposite, synonym);
				
		return synonymComposite;
	}
	
	public Composite createMisappliedName(Taxon misappliedName) {
		
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = getMisappliedGroup();
		
		// Create the name's composite
		Composite composite = new MisappliedNameComposite(this, groupComposite, misappliedName);
		
		return composite;
	}
	
	
	public MisappliedGroupComposite createMisappliedGroup() {
		// Create the group composite
		misappliedGroupComposite = new MisappliedGroupComposite(this, this.getTopLevelComposite());
		misappliedGroupComposite.addFocusListener(new CompositeBorderDecorator(misappliedGroupComposite, this.getManagedForm()));
		
		// Put the group composite before concept group composite, if any
		if (conceptGroupComposite != null) {
			misappliedGroupComposite.moveAbove(conceptGroupComposite);
		}
		
		return misappliedGroupComposite;
	}
	
	
	public Composite createConcept(TaxonRelationship relationship) {	
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = this.getConceptGroup();
		
		// Create the name's composite
		Composite composite = ConceptComposite.getNewInstance(this, groupComposite, relationship);
				
		return composite;
	}

	
	public ConceptGroupComposite createConceptGroup() {
		// Create the group composite
		conceptGroupComposite = new ConceptGroupComposite(this, this.getTopLevelComposite());
		conceptGroupComposite.addFocusListener(new CompositeBorderDecorator(conceptGroupComposite, this.getManagedForm()));
		
		// Put the group composite after misapplied group composite, if any
		if (misappliedGroupComposite != null) {
			conceptGroupComposite.moveBelow(misappliedGroupComposite);
		}
		
		return conceptGroupComposite;

	}

	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.AbstractTaxonEditor#resetGroupComposites()
	 */
	protected void resetGroupComposites() {
		conceptGroupComposite = null;
		misappliedGroupComposite = null;
		homotypicGroupComposites = null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.EditorPart#isDirty()
	 */
	@Override
	public boolean isDirty() {
		return false;
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#dispose()
	 */
	@Override
	public void dispose() {
		super.dispose();
		
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.EditorPart#isSaveAsAllowed()
	 */
	@Override
	public boolean isSaveAsAllowed() {
		return false;
	}
	
	private class NameEditorDropTargetListener extends DropTargetAdapter{
		private GroupedComposite composite;

		public void drop(DropTargetEvent event) {
			Assert.isTrue(event.data instanceof GroupedComposite);
			
			composite = (GroupedComposite) event.data;
			if (composite.getDragSource() != null) {
				// Execute operations at end of drag event
				composite.getDragSource().addListener(DND.DragEnd, new DragListener());
			}
		}	
		
		private class DragListener implements Listener{
			public void handleEvent(Event event) {
				IUndoableOperation operation = null;
				
				// Synonym being dropped
				if(composite instanceof SynonymComposite){
					Synonym synonym = ((SynonymComposite)composite).getSynonym();
					operation = new ChangeHomotypicGroupOperation
							("change type", getUndoContext(), 
									getTaxon(), synonym, HomotypicalGroup.NewInstance(), TaxonNameEditor.this);
				}
				
				// Misapplied name being dropped
				if(composite instanceof MisappliedNameComposite){
						Taxon misapplication = ((MisappliedNameComposite)composite).getMisappliedName();
						operation = new ChangeConceptToSynonymOperation
								("change misapplied name to synonym", getUndoContext(), 
										getTaxon(), misapplication, HomotypicalGroup.NewInstance(), TaxonNameEditor.this);					
				}
				
				// Execute operation if it's been initiated
				if (operation == null) {
					logger.warn("User unsuccessfully tried to drop " + composite.getClass());
				} else {
					EditorUtil.executeOperation(operation);
				}
			}
		}
	}
}