/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.operations;

import java.util.UUID;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;

import eu.etaxonomy.cdm.api.conversation.ConversationHolder;
import eu.etaxonomy.cdm.model.taxon.ITreeNode;
import eu.etaxonomy.cdm.model.taxon.TaxonNode;
import eu.etaxonomy.taxeditor.store.CdmStore;

/**
 * Delete a taxon node from its TaxonomicTree and thus from the current taxonomic view.
 * 
 * 
 * @author n.hoffmann
 * @created 16.01.2009
 * @version 1.0
 */
public class DeleteTaxonNodeOperation extends AbstractPostOperation {
	
	@SuppressWarnings("unused")
	private static Logger logger = Logger.getLogger(DeleteTaxonNodeOperation.class);

	/**
	 * The taxonomical parent of the taxon to be deleted.
	 */
	private ITreeNode parentNode;
	
	public DeleteTaxonNodeOperation(String text, IUndoContext undoContext,
			TaxonNode taxonNode) {
		super(text, undoContext, taxonNode, null);

		parentNode = taxonNode.getParent();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#execute(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus execute(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {

		// Get taxon name - watch out for null names
		String taxonName = "";
		if (taxon.getName() != null) {
			taxonName = taxon.getName().getTitleCache();
		}
		
		// Prompt user for confirmation
		if(! MessageDialog.openConfirm((Shell) info.getAdapter(Shell.class), "Confirm Deletion", "Are you sure you want to delete taxon '" + taxonName + "' from this taxonomic view?")){
			monitor.done();
			return Status.CANCEL_STATUS;
		}
		
		
		// If the taxon node has child nodes, cancel operation
		// TODO add option to continue, and delete children
		if (taxonNode.getChildNodes().size() > 0) {
			MessageDialog.openInformation((Shell) info.getAdapter(Shell.class), "Cannot delete taxon", 
					"'" + taxonName + "' has taxonomic children. " +
					"These must be manually deleted before their parent.");
			monitor.done();
			return Status.CANCEL_STATUS;
		}
		
		
		// delete the taxon in an isolated conversation
		deleteTaxonNodeIsolated(taxonNode.getUuid());
		
		
		return postExecute(null);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#redo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus redo(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {
		return execute(monitor, info);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.operations.AbstractOperation#undo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)
	 */
	@Override
	public IStatus undo(IProgressMonitor monitor, IAdaptable info)
			throws ExecutionException {
		// FIXME we have to add old citation
		parentNode.addChildTaxon(taxon, null, "", null);
		// FIXME readding has to take place in an isolated conversation as well
		// so that all the mediation magic does the rest for us
		return postExecute(null);
	}
	
	/**
	 * 
	 * @param taxonNodeUuid
	 * @return
	 */
	private boolean deleteTaxonNodeIsolated(UUID taxonNodeUuid){
		// get a new conversation
		ConversationHolder conversation = CdmStore.NewTransactionalConversation();
		try{
			TaxonNode taxonNode = CdmStore.getTaxonTreeService().getTaxonNodeByUuid(taxonNodeUuid);
			
			boolean result = taxonNode.delete();
			
			// commit the conversation and throw it away
			conversation.commit(true);
			return result;
		}finally{
			conversation.close();
		}
	}
}
